from enum import Enum
from functools import cached_property
from sympy import Rational

# ========== Atomic Mass ==========
class Element(Enum):
    """
    Atomic weights stored as strings and lazily converted.
    Reference: 
        URL:  https://www.ciaaw.org/atomic-weights.htm
        Year: 2024
    """
    # Period 1
    H = '1.0080'       # Hydrogen
    He = '4.002602'    # Helium
    
    # Period 2
    Li = '6.94'        # Lithium
    Be = '9.0121831'   # Beryllium
    B = '10.81'        # Boron
    C = '12.011'       # Carbon
    N = '14.007'       # Nitrogen
    O = '15.999'       # Oxygen
    F = '18.998403162' # Fluorine
    Ne = '20.1797'     # Neon
    
    # Period 3
    Na = '22.98976928' # Sodium
    Mg = '24.305'      # Magnesium
    Al = '26.9815384'  # Aluminium
    Si = '28.085'      # Silicon
    P = '30.973761998' # Phosphorus
    S = '32.06'        # Sulfur
    Cl = '35.45'       # Chlorine
    Ar = '39.95'       # Argon
    
    # Period 4
    K = '39.0983'     # Potassium
    Ca = '40.078'     # Calcium
    Sc = '44.955907'  # Scandium
    Ti = '47.867'     # Titanium
    V = '50.9415'     # Vanadium
    Cr = '51.9961'    # Chromium
    Mn = '54.938043'  # Manganese
    Fe = '55.845'     # Iron
    Co = '58.933194'  # Cobalt
    Ni = '58.6934'    # Nickel
    Cu = '63.546'     # Copper
    Zn = '65.38'      # Zinc
    Ga = '69.723'     # Gallium
    Ge = '72.630'     # Germanium
    As = '74.921595'  # Arsenic
    Se = '78.971'     # Selenium
    Br = '79.904'     # Bromine
    Kr = '83.798'     # Krypton
    
    # Period 5
    Rb = '85.4678'    # Rubidium
    Sr = '87.62'      # Strontium
    Y = '88.905838'   # Yttrium
    Zr = '91.222'     # Zirconium
    Nb = '92.90637'   # Niobium
    Mo = '95.95'      # Molybdenum
    Ru = '101.07'     # Ruthenium
    Rh = '102.90549'  # Rhodium
    Pd = '106.42'     # Palladium
    Ag = '107.8682'   # Silver
    Cd = '112.414'    # Cadmium
    In = '114.818'    # Indium
    Sn = '118.710'    # Tin
    Sb = '121.760'    # Antimony
    Te = '127.60'     # Tellurium
    I = '126.90447'   # Iodine
    Xe = '131.293'    # Xenon
    
    # Period 6
    Cs = '132.90545196' # Caesium
    Ba = '137.327'      # Barium
    La = '138.90547'    # Lanthanum
    Ce = '140.116'      # Cerium
    Pr = '140.90766'  # Praseodymium
    Nd = '144.242'    # Neodymium
    Sm = '150.36'     # Samarium
    Eu = '151.964'    # Europium
    Gd = '157.249'    # Gadolinium
    Tb = '158.925354' # Terbium
    Dy = '162.500'    # Dysprosium
    Ho = '164.930329' # Holmium
    Er = '167.259'    # Erbium
    Tm = '168.934219' # Thulium
    Yb = '173.045'    # Ytterbium
    Lu = '174.96669'  # Lutetium
    Hf = '178.486'    # Hafnium
    Ta = '180.94788'  # Tantalum
    W = '183.84'      # Tungsten
    Re = '186.207'    # Rhenium
    Os = '190.23'     # Osmium
    Ir = '192.217'    # Iridium
    Pt = '195.084'    # Platinum
    Au = '196.966570' # Gold
    Hg = '200.592'    # Mercury
    Tl = '204.38'     # Thallium
    Pb = '207.2'      # Lead
    Bi = '208.98040'  # Bismuth

    # Period 7
    Th = '232.0377'   # Thorium
    Pa = '231.03588'  # Protactinium
    U = '238.02891'   # Uranium


    @cached_property
    def mass(self):
        return Rational(self.value)



from __future__ import annotations
from pathlib import Path
from importlib import resources
from typing import Iterator
import re

from ..pipeline_contracts import Context, Bundle, LineType, DataLine
from ..errors import InputError, InitHint


patterns = {
    LineType.END:     re.compile(r"^([-_*])\1{2,}$"),         # --- ___ ***
    LineType.COMMENT: re.compile(r"^> +\S"),                  # > comment
    LineType.TITLE:   re.compile(r"^# +(?P<content>.+)"),     # # title
    LineType.REACTION:re.compile(r"^- +(?P<content>\S.+)"),   # - reaction
}

def classify_line(stripped: str) -> DataLine:
    for line_type, pattern in patterns.items():
        match = pattern.match(stripped)
        if match:
            content = match.groupdict().get("content", "")
            return DataLine(line_type, content)
    return DataLine(LineType.NORMAL, stripped)


class InputStage:
    """
    Read the Markdown input file and yield typed line
    """

    def process(self, bundle: Bundle) -> Bundle:
        context = bundle.context
        input_path_name = bundle.stream

        def stream():
            path = Path(input_path_name).expanduser().resolve()
            try:
                with path.open("r", encoding="utf-8") as f:
                    for line in f:
                        stripped = line.strip()
                        if stripped:
                            dataline = classify_line(stripped)
                            if dataline.line_type is LineType.TITLE:
                                if bundle.context.title is None:
                                    bundle.context.title = dataline.data
                            elif dataline.line_type is LineType.COMMENT:
                                continue
                            elif dataline.line_type is LineType.END:
                                break
                            yield dataline
            except FileNotFoundError:
                if input_path_name == "input.md":
                    # Load template from package resources
                    template_text = (
                        resources.files("balancer.templates")
                        .joinpath("input_template.md")
                        .read_text(encoding="utf-8")
                    )

                    # Force the output name to be "input.md" in current working directory
                    target_path = Path("input.md").resolve()
                    target_path.write_text(template_text, encoding="utf-8")

                    
                    raise InitHint(
                        input_path_name,
                        explanation=f"A template has been created at: {target_path}",
                        hint="Please edit this file and run the command again.",
                        )
                else:
                    raise InputError(
                        input_path_name,
                        explanation="This file does not exist in the current working directory.",
                        hint="Check the file path or create the file before running the program.",
                    )
            except IsADirectoryError:
                raise InputError(
                    input_path_name,
                    explanation="This path exists but is a directory, not a file.",
                    hint="Provide a valid file path instead of a directory.",
                )
            except PermissionError:
                raise InputError(
                    input_path_name,
                    explanation="Permission denied when trying to open the file.",
                    hint="Check file permissions or run the program with appropriate privileges.",
                )
        return Bundle(context=context, stream=stream())
    



from __future__ import annotations
from enum import Enum, auto
from typing import TypeVar, Generic, Optional, NamedTuple, Iterable
from dataclasses import dataclass, field
from sympy import Rational, Integer
from openpyxl.styles import Font, Alignment, Border, Side

T = TypeVar("T")

@dataclass(slots=True)
class Context:
    title: Optional(str)=None

class Bundle(NamedTuple):
    context: Context
    stream: T



class LineType(Enum):
    """ All the types of lines in the input file """
    END = auto()
    COMMENT = auto()
    TITLE = auto()
    REACTION = auto()
    NORMAL = auto()

@dataclass(slots=True)
class DataLine:
    line_type: LineType
    data: T



class Substance(NamedTuple):
    formula: str
    info: str
    elements_stoich: dict[str,Rational]
    input_mass: Optional[Rational]


class Reaction(NamedTuple):
    raw: str
    product: Substance
    reactants: list[Substance]
    basis: dict # only keys are used

    def __iter__(self):
        yield from self.reactants
        yield self.product


class SubstanceResult(NamedTuple):
    formula: str
    info: Optional[str]
    mole_mass: Rational
    stoich_in_eq: Integer
    mass: Rational

class ReactionResult(NamedTuple):
    product: SubsatanceResult
    reactants: list[SubstanceResult]



class Style(NamedTuple):
    font: Font
    alignment: Alignment
    border: Border
    
class Cell(NamedTuple):
    value: T
    number_format: str = "@"
    style: Optional[Sytle] = None

class Block(NamedTuple):
    column: Iterable[Cell]
    anchor_row: int
    anchor_col: int
    width: Optional[float]=None



from __future__ import annotations
from typing import Protocol, TypeVar, Generic, Sequence
from .pipeline_contracts import Context, Bundle

InT = TypeVar("InT")
OutT = TypeVar("OutT")

class Stage(Protocol):
    def process(self, bundle: Bundle) -> Bundle:
        ...


class Pipeline(Generic[InT, OutT]):
    """Executes a series of stages in sequence."""
    def __init__(self, *stages: Sequence[Stage]):
        self.stages = stages

    def run(self, data: InT) -> OutT:
        bundle = Bundle(context=Context(), stream=data)

        for stage in self.stages:
            bundle = stage.process(bundle)
        return bundle




import argparse
from balancer import Stage, Pipeline
from balancer import InputStage, ParseStage, ComputeStage, FormatStage, OutputStage

def main()->None:
    parser = argparse.ArgumentParser(description="Balancer 2.0 — Chemical equation balancer")
    parser.add_argument(
        "input_file",
        nargs="?",
        default="input.md",
        help="Path to the input Markdown file (default: input.md)",
    )
    args = parser.parse_args()

    Pipeline(
        InputStage(),
        ParseStage(),
        ComputeStage(),
        FormatStage(),
        OutputStage(),
    ).run(args.input_file)


def wait_for_any_key():
    try:
        import msvcrt # for windows
        print("Press any key to exit...", end="", flush=True)
        msvcrt.getch()  
    except ImportError:
        input("Press Enter to exit...")  # fallback

if __name__ == "__main__":
    main()
    wait_for_any_key()


if __name__ == "__main__":
    main()
    wait_for_any_key()



from __future__ import annotations
from typing import Optional, Iterator, Iterable
from functools import reduce
from sympy import Matrix, Rational, Integer, lcm, gcd, symbols, linsolve, S, Interval, Union
from sympy.solvers.inequalities import reduce_rational_inequalities

from ..atomic_mass import Element
from ..pipeline_contracts import Context,Bundle,LineType, DataLine, Substance, SubstanceResult, Reaction, ReactionResult
from ..errors import ComputeError

# ======== Compute ======== 
def to_vector(basis:dict[str,Optional[Rational]],elements_stoich:dict[str,Rational])->Matrix:
    return Matrix([elements_stoich.get(e, Rational(0)) for e in basis])

def non_negative(stoichs:Iterable[Rational])->bool:
    for stoich in stoichs:
        if stoich < 0:
            return False
    return True

def solve_reaction(reactant_matrix:Matrix, product_vector:Matrix, reaction:Reaction)->list[Rational]:
    x = symbols(f'x0:{reactant_matrix.cols}')
    sol_set = linsolve((reactant_matrix, product_vector), x)
    sol_non_negative = True

    if sol_set:
        sol = list(sol_set.args[0])
        if not sol_set.free_symbols:
            sol_non_negative = non_negative(sol)

        elif len(sol_set.free_symbols) == 1:
            (free_symbol,) = tuple(sol_set.free_symbols)
            inequalities = [expr>=0 for expr in sol]
            iq_sol_set = reduce_rational_inequalities([inequalities], free_symbol)

            if iq_sol_set in (S.EmptySet, S.false, False):
                # S.EmptySet: solution set is empty
                # S.False:    False after bool reduce
                sol = [expr.subs(free_symbol, S.Zero) for expr in sol]
                sol_non_negative = False
            else:
                # Find a simple solution in the set
                if hasattr(iq_sol_set, "as_set"):
                    iq_sol_set = iq_sol_set.as_set()
                
                if isinstance(iq_sol_set, Interval):
                    infimum = iq_sol_set.inf
                # Defensive programming
                elif isinstance(iq_sol_set, Union):
                    candidates = []
                    for iv in iq_sol_set.args:
                        if iv.inf is S.NegativeInfinity:
                            if iv.sup == S.Infinity:
                                candidates.append(S.Zero)
                            else:
                                candidates.append(iv.sup)
                        else:
                            candidates.append(iv.inf)
                    infimum = min(infs)
                # Defensive programming * 2
                else:
                    #fallback to zero
                    infimum = S.Zero
                    sol_non_negative = False

                sol = [expr.subs(free_symbol, infimum) for expr in sol]
                    

        else:
            reactant_formulas = (reactant.formula for reactant in reaction.reactants)
            raise ComputeError(
                    explanation="Too many unnessary reactants, which lead to more than 1 free variables in the solution. Please remove some from these:",
                    hint=", ".join(reactant_formulas)
                    )

        sol.append(Rational(1))
        if sol_non_negative:
            return sol
        else:
            sol = normalize_ratio(sol)
            for i,stoich in enumerate(sol):
                if stoich == 1:
                    sol[i] = ""
                elif stoich == -1:
                    sol[i] = "-"
                
            reactant_formulas = [reactant.formula for reactant in reaction.reactants]
            product_formula = reaction.product.formula
            raise ComputeError(
                    explanation="No non-negative solution. A solution is:",
                    hint=" ".join((
                        f"{sol[0]}{reactant_formulas[0]}",
                        *((f"+ {stoich}{formula}").replace("+ -", "- ") for stoich,formula in zip(sol[1:-1],reactant_formulas[1:])),
                        "===",
                        f"{sol[-1]}{product_formula}"
                        )
                    ))

    else:
        # just no solution
        # maybe the user writed wrong Element symbols?
        basis_elements = set(reaction.basis)
        reactant_elements = set().union(
                *(reactant.elements_stoich for reactant in reaction.reactants)
                )
        product_elements = set(reaction.product.elements_stoich)
        
        elements_not_in_reactant = basis_elements - reactant_elements
        elements_not_in_product = basis_elements - product_elements
        if elements_not_in_reactant:
            raise ComputeError(
                explanation="These required elements are missing from the reactants:",
                hint=", ".join(elements_not_in_reactant)
            )
        if elements_not_in_product:
             raise ComputeError(
                explanation="These required elements are missing from the product:",
                hint=", ".join(elements_not_in_product)
            )           

        raise ComputeError(
            hint="This chemical equation has no valid stoichiometric solution.",
        )

        

    
def normalize_ratio(stoichs:list(Rational))->list(Integer):
    """
        lcm: least common multiple
        gcd: greast common divisor
        p: numerator
        q: denominator
    """
    lcm_of_denominator = reduce(lcm,
                                [n.q for n in stoichs if n!=0]
                                )
    integerized = [n * lcm_of_denominator for n in stoichs]
    gcd_of_integers = reduce(gcd,integerized)
    return [n / gcd_of_integers for n in integerized]

def mole_mass(elements_stoich:dict[str,Rational])->Rational:
    try:
        return sum([Rational(Element[e].value) * n for e,n in elements_stoich.items()])
    except KeyError as e:
            raise ComputeError(
                explanation=f"Unknown element symbol encountered:",
                hint=e.args[0]
            )

def mass(stoichs:list, mole_masses:list, input_masses:iterator[Optional[Rational]])->list[Rational]:
    for i,input_mass in enumerate(input_masses):
        if input_mass is not None:
            eq_mole = input_mass / mole_masses[i] / stoichs[i]
            break
    else:
        raise ComputeError(
            explanation="No input mass was provided for any reactant or product.",
            hint="Please write mass in grams (g) near the product or one of the reactants."
            )

    return [eq_mole * stoich * mole_mass for stoich,mole_mass in zip(stoichs,mole_masses)]




def balancer(reaction:Reaction)->Optional[ReactionResult]:
    basis = reaction.basis
    reactant_cnt = len(reaction.reactants)
    reactant_matrix = Matrix.hstack(*[to_vector(basis, reactant.elements_stoich) for reactant in reaction.reactants])
    product_vector = to_vector(basis, reaction.product.elements_stoich)
    reactant_product_stoich = solve_reaction(
            reactant_matrix,
            product_vector,
            reaction
            )

    reactant_product_stoich=normalize_ratio(reactant_product_stoich)
    reactant_product_mole_mass = [mole_mass(s.elements_stoich) for s in reaction]
    reactant_product_mass = mass(
                                 reactant_product_stoich,
                                 reactant_product_mole_mass,
                                 (substance.input_mass for substance in reaction),
                                )

    # ok!
    *reactants_mole_mass, product_mole_mass = reactant_product_mole_mass
    *reactants_stoich, product_stoich = reactant_product_stoich
    *reactants_mass, product_mass = reactant_product_mass

    return ReactionResult(
            SubstanceResult(
                reaction.product.formula,
                reaction.product.info,
                product_mole_mass,
                product_stoich,
                product_mass,
                ),
            [
                SubstanceResult(
                    reactant.formula,
                    reactant.info,
                    mole_mass,
                    stoich,
                    mass)
                for reactant,mole_mass,stoich,mass
                in zip(
                    reaction.reactants,
                    reactants_mole_mass,
                    reactants_stoich,
                    reactants_mass
                    )
                ],
            )

    
    
# ======== Stage ========
class ComputeStage:
    """ balance the reaction """
    def process(self, bundle: Bundle)->Bundle:
        datalines=bundle.stream

        def stream():
            for dataline in datalines:
                if dataline.line_type is LineType.REACTION:
                    raw_reaction = dataline.data.raw
                    try:
                        dataline.data = balancer(dataline.data)
                    except ComputeError as e:
                        e = ComputeError(
                                raw_reaction,
                                explanation=e.explanation,
                                hint=e.hint,
                                )
                        print(e)
                        continue
                yield dataline

        return Bundle(context=bundle.context, stream=stream())
